// AmlQuotes.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include <stdio.h> //sprintf declaration
#include "APPHelper.h" //Plugin API Service functions declarations

static BOOL m_bEnabled=FALSE;//enabled auto completion
static HMODULE m_hDLL=NULL;//need for ReReadPluginInfo function

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  reason, 
                       LPVOID lpReserved)

{
	if (DLL_PROCESS_ATTACH==reason)
		m_hDLL=(HMODULE)hModule;
    return TRUE;
}

//registry key paths for options
#ifndef _DEBUG//for debug configuration
	#define KEY_PATH "Software\\G&G Software\\Aml Pages\\Plugins\\Aml Quotes"
#else//for release configuration
	#define KEY_PATH "Software\\G&G Software\\Aml Pages\\Plugins\\Aml Quotes Debug"
#endif//#ifndef _DEBUG

#define ENABLED "AutoQuotesEnabled"

//save plugin options
void WriteOptions()
{
	HKEY hkey=NULL;
	if (ERROR_SUCCESS==::RegCreateKey(HKEY_CURRENT_USER,KEY_PATH,&hkey)) {
		::RegSetValueEx(hkey,ENABLED,0,REG_DWORD,(BYTE*)&m_bEnabled,sizeof(m_bEnabled));
		RegCloseKey(hkey);
	}
}

//read plugin options
BOOL ReadOptions()
{
	HKEY hkey=NULL;
	if (ERROR_SUCCESS==RegOpenKeyEx(HKEY_CURRENT_USER,KEY_PATH,0,KEY_QUERY_VALUE,&hkey)) {
		DWORD dwType=REG_DWORD;
		DWORD dwSize=sizeof(m_bEnabled);
		::RegQueryValueEx(hkey,ENABLED,NULL,&dwType,(LPBYTE)&m_bEnabled,&dwSize);
		RegCloseKey(hkey);
	}
	else 
		m_bEnabled=TRUE;
		
	return m_bEnabled;
}

//Get Information About Plugin
extern "C"
BOOL __declspec(dllexport) Aml_GetPluginInfo(AML_PLUGIN_INFO* p)
{
#ifdef _DEBUG//debug test only
	if (p->cbSize>=SIZE_AML_PLUGIN_INFO_VER2) {//member wLanguage exist in AML_PLUGIN_INFO
		char sz[100]={0};
		sprintf(sz,"Aml Pages Interface Use Language ID=%d\n",p->wLanguage);
		OutputDebugString(sz);
		/*
		if (p->wLanguage==LANG_RUSSIAN)
			...
		else if (p->wLanguage==LANG_ENGLISH)
			.,,
		else
			(p->wLanguage==LANG_NEUTRAL)
		*/
	}
#endif
	ReadOptions();//read plugin options
	if (m_bEnabled)//if enabled, subscribe events of Aml Pages
		p->dwEventMask=APFEM_EDITOREVENTS;//events of text editor
	strcpy(p->szMenuName,"Aml Quotes (Sample)");//popup menu name
	p->nMenuCount=1;//menu command count
	return TRUE;//ok
}

extern "C"
BOOL __declspec(dllexport) Aml_GetMenuItemInfo(UINT nIndex, AML_MENU_ITEM_INFO* p)
{
	BOOL bRet=TRUE;
	switch(nIndex) {//first menu command
	case 0:
		strcpy(p->szMenuName,"Quotes Auto Completion");//menu text
		strcpy(p->szFunctionName,"EnableProc");//Plugin Procedure (exportable name)
		strcpy(p->szSBHint,"Automatically insert single or double back quote");//hint in status bar
		p->nPluginType=APPT_SHOWDIALOG;//Plugin Procedure Type
		if (m_bEnabled) {//auto completion enabled?
			strcat(p->szSBHint,TEXT(" (enabled)"));
			p->nMenuState=MF_CHECKED;//checked menu
		}
		else//or auto completion disabled?
			strcat(p->szSBHint,TEXT(" (disabled)"));
		break;
	default://??? see above (Aml_GetPluginInfo - >nMenuCount=1)
		bRet=FALSE;
	};//switch(nIndex) {

	return bRet;//OK
}

//on user click in plugin menu
extern "C"
BOOL __declspec(dllexport) EnableProc(HWND)
{
	m_bEnabled=!m_bEnabled;//change options
	WriteOptions();//write plugin options
	APPHelper::ReReadPluginsInfo(m_hDLL);//remake plugin menu
	return TRUE;//OK
}

bool IsLeftRightSymbol(const HWND hwndEdit
					   , const UINT nChar
					   , const LONG lPos)
{
	TEXTRANGE tr={0};
	tr.chrg.cpMin=lPos+1;
	tr.chrg.cpMax=tr.chrg.cpMin+1;
	char szText[2]={0};
	tr.lpstrText=szText;
	const LRESULT l=SendMessage(hwndEdit,EM_GETTEXTRANGE,0,(LPARAM)&tr);
	if ((l>0) && (nChar==(UINT)szText[0]))
		return true;

	if (lPos>0) {
		szText[0]=0;
		tr.chrg.cpMin=lPos-1;
		tr.chrg.cpMax=tr.chrg.cpMin+1;
		const LRESULT l=SendMessage(hwndEdit,EM_GETTEXTRANGE,0,(LPARAM)&tr);
		if ((l>0) && (nChar==(UINT)szText[0]))
			return true;
	}//if (lPos>0) {

	return false;
}

struct PARAM {
	const HWND hwnd;//hwnd of Text Editor
	const UINT nChar;//key pressed
	const LONG lCaretPos;//caret position
	PARAM(HWND h, UINT c, LONG caretPos)
		:hwnd(h)
		,nChar(c)
		,lCaretPos(caretPos)
	{}
};//struct PARAM {


DWORD __stdcall ChangeProc(LPVOID p)
{
	PARAM* const pParam=(PARAM*)p;
		char sz[2]={0};
		//insert back quotes (single or double)
		sz[0]=pParam->nChar;
		SendMessage(pParam->hwnd,EM_REPLACESEL,FALSE,(LPARAM)(LPCTSTR)sz);
		
		//get current selectio after replace back quotes
		CHARRANGE cr={0};
		SendMessage(pParam->hwnd,EM_EXGETSEL,0,(LPARAM)&cr);
		cr.cpMin--;
		cr.cpMax=cr.cpMin;
		//set selection between forward and back quotes
		SendMessage(pParam->hwnd,EM_EXSETSEL,0,(LPARAM)&cr);
	delete pParam;
	return 0;
}

#define Q_SINGLE 39//single quote
#define Q_DOUBLE 34//double quote

void Process(const HWND hwndEdit/*hwnd of Text Editor*/
					,const UINT nChar/*char pressed*/
					,const CHARRANGE& cr/*caret position*/)
{
	//this is debug test
	//char sz[2]={0};
	//sz[0]=nChar;
	//char res[10]={0};
	//sprintf(res,"%s %d\n",sz,nChar);
	//OutputDebugString(res);


	if ((Q_SINGLE==nChar) || (Q_DOUBLE==nChar)) {//if single or double quotes pressed
		if (!IsLeftRightSymbol(hwndEdit,nChar,cr.cpMin)) {//check left of right quotes exist?
			//save HWND of Text Editor, char and caret pos
			const PARAM* const pParam=new PARAM(hwndEdit,nChar,cr.cpMin);
			if (!IsBadReadPtr(pParam,sizeof(PARAM))) {
				DWORD id;
				//make new thread
				const HANDLE h=CreateThread(NULL,0,ChangeProc,(LPVOID)pParam,0,&id);
				if (h)//if thread is created ??
					CloseHandle(h);//close handle of thread
			}//if (!IsBadReadPtr(pParam,sizeof(PARAM)))
		}//if (!IsLeftRightSymbol(hwndEdit,nChar,cr.cpMin)) {
	}//if ((Q_SINGLE==nChar) || (Q_DOUBLE==nChar)) {
}

//on Aml Pages Event
extern "C"
BOOL __declspec(dllexport) Aml_EventHandler(AML_EVENTINFO* p)
{
	switch(p->EventCode/*event type code*/) {
	case AEC_EDITOR://event of text editor
		switch(p->EventInfo.EditorEvent.nMessage) {
		case WM_CHAR://WM_CHAR window message
			//call out procedure
			Process(p->EventInfo.EditorEvent.hwnd,p->EventInfo.EditorEvent.nChar,p->EventInfo.EditorEvent.crSel);
			break;
		};//switch(p->EditorEvent.nMessage) {
		break;
	}//switch(p->EventCode) {

	return TRUE;//OK
}